/*  $Id: Queue.hh,v 1.5 1999/07/16 20:18:48 shaggy Exp $
 *      relative path: ./lib/c++
 */

#ifndef __Queue_hh__
#define __Queue_hh__

class QueueNode {
	// friend class Queue;
	QueueNode *prev;
	QueueNode *next;
public:
	void clear() { prev = next = this; }
	bool empty() const { return next == this; }
	QueueNode& insert(QueueNode *node) {
			node->prev = prev; node->next = this;
			prev->next = node; prev = node; return *node; }
	QueueNode& append(QueueNode *node) {
			node->next = next; node->prev = this;
			next->prev = node; next = node; return *node; }
	QueueNode& get_next() const { return *next; }
	QueueNode& get_prev() const { return *prev; }
	QueueNode& remove() {
			next->prev = prev;
			prev->next = next; return *this; }
	QueueNode& rotate_next() {
			QueueNode *node = next;
			next = node->next;
			node->next->prev = this;
			prev->next = node;
			node->prev = prev;
			prev = node;
			node->next = this; return *node; }
	QueueNode& rotate_prev() {
			QueueNode *node = prev;
			prev = node->prev;
			node->prev->next = this;
			next->prev = node;
			node->next = next;
			next = node;
			node->prev = this; return *node; }
};


template<class T> class QueueIterator {
	struct Node : QueueNode {
		T item;
		Node(const T &newitem) { item = newitem; }
	} *current;
public:
	QueueIterator(const QueueNode *c) { current = (Node*)c; }
	T& operator* () { return current->item; }
	operator QueueNode() { return *current; }
	QueueIterator& operator++ () { current = (Node*)&current->get_next(); return *this; }
	QueueIterator& operator-- () { current = (Node*)&current->get_prev(); return *this; }
	QueueIterator operator++ (int) { Node *old = current;
		current = (Node*)&current->get_next(); return old; }
	QueueIterator operator-- (int) { Node *old = current;
		current = (Node*)&current->get_prev(); return old; }
	QueueIterator insert(const T &item) { return (Node*)&current->insert(new Node(item)); }
	QueueIterator append(const T &item) { return (Node*)&current->append(new Node(item)); }
	QueueIterator& remove() { Node *old = current; operator++();
			delete (Node*)(&old->remove()); return *this; }
	bool operator== (const QueueIterator &q) const
		{ return current == q.current; }
	bool operator!= (const QueueIterator &q) const
		{ return current != q.current; }
};



template<class T> class Queue {
	friend class QueueIterator<T>;
	QueueNode queue;
public:
	Queue() { queue.clear(); }
	bool empty() const { return queue.empty(); }
	QueueIterator<T> head() const { return QueueIterator<T>(&queue.get_next()); }
	QueueIterator<T> tail() const { return QueueIterator<T>(&queue.get_prev()); }
	QueueIterator<T> begin() const { return QueueIterator<T>(&queue.get_next()); }
	QueueIterator<T> end() const { return QueueIterator<T>(&queue); }
	QueueIterator<T> rotate_next() { return QueueIterator<T>(&queue.rotate_next()); }
	QueueIterator<T> rotate_prev() { return QueueIterator<T>(&queue.rotate_prev()); }
	~Queue() { while(!queue.empty()) head().remove(); }
};

#endif
