/*  $Id: atoi.cc,v 1.10 2000/01/12 03:44:45 shaggy Exp $
 *      convert an ascii string to a number
 *      relative path: lib/posix
 *  
 *  Copyright (c) 1994 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/members/shaggy
 */

#include <stdx1.h>
#include <stddef.h>
#include "stdlib.h"

#define digit_value(ch)  (ch<='9'?ch-'0':10+(ch>='A'&&ch<='Z'?ch-'A':ch-'a'))


/* numbers are accepted in the following format:
 *   ddddd   default base
 *   0dddd   octal
 *   0xddd   hex
 *   0tddd   decimal
 *   0nddd   binary
 */

ulong
strtoul(const char *str, char **adr, int base)
{
	int x;
	ulong num = 0;
	if(!base)
	base = 10;
	while(*str == ' ' || *str == '\f' || *str == '\n'
			|| *str == '\r' || *str == '\t' || *str == '\v')
		str++;
	if(*str == '0' && base == 0) {
		base = 8;
		str++;
		switch(*str) {
		case 'x': case 'X':
			base = 16; str++; break;
		case 'b': case 'B':
			base = 2; str++; break;
		}
	}
	while((x = digit_value(*str)) >= 0 && (x < base)) {
		num = (num*base)+x;
		str++;
	}
	if(adr)
		*adr = (char*)str;
	return num;
}


long
strtol(const char *str, char **adr, int base)
{
	if(*str == '-')
		return -(long)strtoul(&str[1], adr, base);
	else
		return (long)strtoul(str, adr, base);
}


int
atoi(const char *str)
{
	return (long)strtol(str,0,10);
}

long
atol(const char *str)
{
	return (long)strtol(str,0,10);
}

ulong
atoul(const char *str)
{
	return (ulong)strtoul(str,0,10);
}


uchar
bcd_decode(uchar byte)
{
	return ((byte>>4)&0xf)*10 + (byte&0xf);
}

uchar
bcd_encode(uchar byte)
{
	return ((byte/10)<<4) | (byte%10);
}

