/*  $Id: memcpy.cc,v 1.3 2000/01/12 03:44:45 shaggy Exp $
 *      string manipulation functions
 *      relative path: lib/posix
 *  
 *  Copyright (c) 1994 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/members/shaggy
 */

#include <stdx1.h>
#include "stddef.h"
#include "string.h"

void*
memcpy(void *s1, const void *s2, size_t len)
{
	register ulong *src, *dest;
	register byte *b1, *b2;
	b1 = (byte*)s1;
	b2 = (byte*)s2;
	while(len && ((vaddr_t)b1&3)) {
		*b1++ = *b2++;
		len--;
	}
	src = (ulong*)b2;
	dest = (ulong*)b1;
	while(len>16) {
		*dest++ = *src++;
		*dest++ = *src++;
		*dest++ = *src++;
		*dest++ = *src++;
		len -= 16;
	}
	b1 = (byte*)dest;
	b2 = (byte*)src;
	while(len) {
		*b1++ = *b2++;
		len--;
	}
	return s1;
}

void*
memmove(void *s1, const void *s2, size_t len)
{
	if(s1 <= s2)
		return memcpy(s1, s2, len);
	else {
		register byte *b1 = (byte*)s1+len;
		register byte *b2 = (byte*)s2+len;
		for(;len>0; len--) 
			*--b1 = *--b2;
		return s1;
	}
}

void*
memset(void *s, int val, size_t len)
{
	register ulong *dest;
	register ulong lfill;
	register byte *adr;
	adr = (byte*)s;
	lfill = (val)|(val<<8)|(val<<16)|(val<<24);
	while(len && ((vaddr_t)adr&3)) {
		*adr++ = val;
		len--;
	}
	dest = (ulong*)adr;
	while(len>16) {
		*dest++ = lfill;
		*dest++ = lfill;
		*dest++ = lfill;
		*dest++ = lfill;
		len -= 16;
	}
	adr = (byte*)dest;
	while(len--)
		*adr++ = val;
	return s;
}

void*
memsetw(void *s, ushort val, size_t len)
{
	register ushort *dest;
	dest = (ushort*)s;
	while(len-- > 0)
		*dest++ = val;
	return s;
}

void*
memsetl(void *s, ulong val, size_t len)
{
	register ulong *dest;
	dest = (ulong*)s;
	while(len-- > 0)
		*dest++ = val;
	return s;
}

