/*  $Id: stdx1.h,v 1.2 1999/07/16 20:18:48 shaggy Exp $
 *	"standard" extensions to the C language, version 1
 *	types, constants, and macros that are to be made available to
 *	all C/C++ programs in this source tree
 *      relative path: lib
 */ 

#ifndef __stdx1_h__
#define __stdx1_h__

#ifdef __cplusplus
#	define extern_C	extern "C"
#	define extern_C_begin	extern "C" {
#	define extern_C_end	}
#else
#	define extern_C	extern
#	define extern_C_begin
#	define extern_C_end
#endif

#if 0
#if !defined(__cplusplus) || defined(need_wchar_t)
	typedef ushort		wchar_t;
#endif
#endif

#include <inttypes.h>
#include <ltypes.h>

#if 0	/* Not needed unless inttypes.h is switched to use '_t' extensions */
typedef int8_t		int8;
typedef int16_t		int16;
typedef int32_t		int32;
typedef int64_t		int64;
typedef uint8_t		uint8;
typedef uint16_t	uint16;
typedef uint32_t	uint32;
typedef uint64_t	uint64;
#endif

/* okay, let's get this over with once and for all, and start calling
 * these things what they are... not unsigned chars, uchars, uint8's,
 * and especially not chars or signed chars...  When we're dealing
 * with chunks of memory as simply streams of bytes, this should
 * be made obvious:
 */
typedef uint8 byte;

/* uchar and schar should both be eliminated-- they are used for 8-bit 
 * numerical quantities, which have nothing to do with characters except
 * that the storage size is the same-- use byte, uint8, or int8 instead.
 */ 
typedef unsigned char		uchar;	/* should use byte or uint8 instead */
typedef signed char		schar;  /* should use int8 instead */
typedef unsigned short int	ushort;
typedef unsigned long int	ulong;
typedef unsigned int		uint;
typedef unsigned long long int	uvlong;
typedef long long int		vlong;
 
typedef uint8	flag8;
typedef uint16	flag16;
typedef uint32	flag32;
typedef uint64	flag64;

/* Strictly speaking, these aren't necessary, but they aid in 
 * the clarity and conciseness of code.  `char*' by itself is fairly
 * short, but `const char *' starts getting rather large, and is used
 * quite commonly.  In addition, `char*' doesn't make it clear whether
 * it is a pointer to a single character or an array of characters, and
 * in no way shows whether that array is supposed to be zero terminated.
 * The following types make it clear for zero terminated strings:
 * (Note that the pointer was purposely NOT included in these typedefs,
 * so that when these types are used, an explicit '*' is still required,
 * making it very clear that they aren't self-contained entities, but
 * instead that they reference memory elsewhere, which may need to be
 * explicitly managed)
 */
typedef char sz;
typedef const char csz;
#if 0
typedef char msz;		/* multibyte (UTF8?) strings */
typedef const char mcsz;
typedef wchar_t wsz;		/* wide strings */
typedef const wchar_t wcsz;
#endif

#define _K_	1024			// KiloBytes (2^10)
#define	_M_	1048576			// MegaBytes (2^20)
#define _G_	1073741824		// GigaBytes (2^30)
#define _T_	1099511627776LL		// TeraBytes (2^40)
#define _P_	1125899906842624LL	// PetaBytes (2^50)
#define _E_	1180591620717411303424LL// ExaBytes (2^60)
/* anything beyond this (Zetta, Yotta, ?) won't fit in a 64-bit integer,
 * so there's no point in adding them until larger ints become available
 */

/* 'nil' looks so much nicer than 'NULL' */
#ifdef __cplusplus
#	define nil	0
#else
#	define nil	((void*)0L)
#endif

#ifndef __cplusplus
	typedef byte bool;
	enum { true = 1, false = 0 };
#endif

/* number of elements in an array: */
#define nelem(a)	(sizeof(a)/sizeof(*(a)))

/* for evaluating and quoting within macros: */
#define QUOTE(n)	__quote_(n)
#define __quote_(x)	#x


/* the rest of this file still needs cleaning up */

#define align(adr, n)		(((adr)+(n)-1) & ~((n)-1))
#define align_down(adr, n)	((adr) & ~((n)-1))
#define align_div(adr, n)	(((adr)+(n)-1)/(n))

#define is_pow2(n)	(!((n) & ((n)-1))

/* abs comes from stdlib... */
/*
#define abs(x)		(((x) < 0) ? (-(x)) : (x))
*/

#define min(x,y)	(((x) < (y)) ? (x) : (y))
#define max(x,y)	(((x) > (y)) ? (x) : (y))

/*
#define ADR_LEN(d)	&d, sizeof(d)
#define CSTR_LEN(t,v)	(t)(v), sizeof(v)-1
#define DSTR_LEN(t,v)	(t)(v), strlen(v)
*/

#define range_min(v,min)	 { if((v) < (min)) (v) = (min); }
#define range_max(v,max)	 { if((v) > (max)) (v) = (max); }
#define range_min_max(v,min,max) { if((v) < (min)) (v) = (min); \
				   if((v) > (max)) (v) = (max); }

/* things we don't know what to do with yet: */

extern_C_begin 

int eprintf(csz *fmt, ...);
void *memsetw(void *adr, ushort v, _size_t n);
void *memsetl(void *adr, ulong v, _size_t n);

extern_C_end

#endif
