/*  $Id: fdisk.cc,v 1.27 2000/01/17 05:03:56 shaggy Exp $
 *      
 *      Fixed Disk Partitioning Utility
 *      
 *      relative path: sys/boot/m-pc
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/members/shaggy
 */

#define fdisk_VER_MAJOR	0
#define fdisk_VER_MINOR	3
#define fdisk_VER_PATCH	0
#define fdisk_VER_STR	"0.3"
#define fdisk_VER_DATE	"January 16, 2000"
#define fdisk_VER_NDATE	"01/16/00"


#include <libc_posix.h>
#include <stdx1.h>
#include <defines>
#include <boot/libsolo.h>
#include <boot/fileinfo.h>
#include <boot/boot.h>
#include <hw/Port.h>


static csz rcsid[] __attribute__((unused))
	= "$Id: fdisk.cc,v 1.27 2000/01/17 05:03:56 shaggy Exp $"
	  "$Release: " fdisk_VER_STR " $";

//#define USE_DEV_PORT


struct PartitionType {
	uint8 id;
	csz *name;
};


// Hard-coded for now.... later to be read from a configurable file...

PartitionType ptype[] = {
	{0x00, "Empty"},
	{0x01, "DOS 12-bit FAT"},
	{0x02, "XENIX root"},
	{0x03, "XENIX usr"},
	{0x04, "DOS 16-bit <32M"},
	{0x05, "Extended"},
	{0x06, "DOS 16-bit >=32M"},
	{0x07, "HPFS/NTFS"},
	{0x08, "AIX"},
	{0x09, "AIX bootable"},
	{0x0a, "OS/2 Boot Manager"},
	{0x0b, "Win95 FAT32"},
	{0x0c, "Win95 FAT32 (LBA)"},
	{0x0e, "Win95 FAT16 (LBA)"},
	{0x0f, "Win95 Extended (LBA)"},
	{0x10, "OPUS"},
	{0x11, "Hidden FAT12"},
	{0x12, "Compaq diagnostic"},
	{0x14, "Hidden FAT16 < 32M"},
	{0x16, "Hidden FAT16"},
	{0x17, "Hidden HPFS/NTFS"},
	{0x18, "AST Windows swap"},
	{0x1b, "Hidden Win95 FAT32"},
	{0x1c, "Hidden Win95 FAT32 (LBA)"},
	{0x1e, "Hidden Win95 FAT16 (LBA)"},
	{0x24, "NEC DOS"},
	{0x3c, "PartitionMagic"},
	{0x40, "Venix 80286"},
	{0x41, "PPC PReP Boot"},
	{0x42, "SFS"},
	{0x4d, "QNX4.x"},
	{0x4e, "QNX4.x 2nd part"},
	{0x4f, "QNX4.x 3rd part"},
	{0x50, "OnTrack DM"},
	{0x51, "OnTrack DM6 Aux"},
	{0x52, "CP/M"},
	{0x53, "OnTrack DM6 Aux"},
	{0x54, "OnTrackDM6"},
	{0x55, "EZ-Drive"},
	{0x56, "Golden Bow"},
	{0x5c, "Priam Edisk"},
	{0x61, "SpeedStor"},
	{0x63, "GNU HURD"},
	{0x64, "Novell Netware"},
	{0x65, "Novell Netware"},
	{0x70, "DiskSecure Mult"},
	{0x75, "PC/IX"},
	{0x80, "Old MINIX"},
	{0x81, "MINIX/Old Linux"},
	{0x82, "Linux swap"},
	{0x83, "Linux native"},
	{0x84, "OS/2 hidden C:"},
	{0x85, "Linux extended"},
	{0x86, "NTFS volume set"},
	{0x87, "NTFS volume set"},
	{0x93, "Amoeba"},
	{0x94, "Amoeba BBT"},
	{0xa5, "BSD/386"},
	{0xa6, "OpenBSD"},
	{0xa7, "NEXTSTEP"},
	{0xae, "ShagOS"},		/* Note: these are new and */
	{0xaf, "ShagOS swap"},		/* need to be properly reserved */
	{0xb7, "BSDI fs"},
	{0xb8, "BSDI swap"},
	{0xc1, "DRDOS/sec"},
	{0xc4, "DRDOS/sec"},
	{0xc6, "DRDOS/sec"},
	{0xc7, "Syrinx"},
	{0xdb, "CP/M"},
	{0xe1, "DOS access"},
	{0xe3, "DOS R/O"},
	{0xe4, "SpeedStor"},
	{0xeb, "BeOS fs"},
	{0xf1, "SpeedStor"},
	{0xf2, "DOS secondary"},
	{0xf4, "SpeedStor"},
	{0xfd, "Linux raid"},
	{0xfe, "LANstep"},
	{0xff, "BBT"}
};



// Partition table notes:
// Physical sector/head/cylinder are always absolute
// Logical start is absolute for primary partitions
// For the extended partition, the 'bootsector' contains another
// partition table.
//	The first entry is relative to the main
// 	extended partition, and is the first logical partition
// 	The second entry is another extended entry (these act
//	like a linked list)
//	from here on, each extended entry has its logical address
//	relative to the main extended partition,
//	and each individual partition entry has its logical address
//	relative to the extended partition in which it is found 
 

int cylinders, heads, sectors;



int
getint(int default_val)
{
	sz buf[40];

	if(!gets(buf, sizeof(buf)) || buf[0] < '0' || buf[0] > '9')
		return default_val;
	return atoi(buf);
}


csz*
ptype_name(uint8 id)
{
	for(size_t i=0; i<nelem(ptype); i++)
		if(ptype[i].id == id)
			return ptype[i].name;
	return "unknown";
}




void
chs_unpack(uint &cyl, uint &/*head*/, uint &sec)
{
	uint newcyl = cyl+((sec&0xc0)<<2);
	sec &= 0x3f;
	cyl = newcyl;
}


int
chs_to_linear(uint cyl, uint head, uint sec)
{
	return ((cyl*heads)+head)*sectors+(sec-1);
}

int
pchs_to_linear(uint cyl, uint head, uint sec)
{
	chs_unpack(cyl, head, sec);
	return chs_to_linear(cyl, head, sec);
}



int
partition_start(Partition *part, int ofs = 0)
{
	int start = part->start + ofs;
	int start2 = pchs_to_linear(part->startcylinder,
			part->starthead, part->startsector);
	if(start != start2) {
		int count1024 = (start - start2)/(1024*heads*sectors);
		if(count1024 > 0) {
			start2 = pchs_to_linear(
				part->startcylinder+count1024*1024,
				part->starthead, part->startsector);
			printf("Warning: start cylinder invalid (cyl >= 1024)\n");
		}	
		if(start != start2)
			printf("Warning: inconsistent start = %d+%d = %d, chs -> %d\n",
				(int)part->start, ofs, start, start2);
	}
	return start;
}


int
partition_size(Partition *part, int ofs = 0)
{
	int end = part->start+ofs+part->size-1;
	int end2 = pchs_to_linear(part->endcylinder,
			part->endhead, part->endsector);
 
	if(end != end2) {
		int count1024 = (end-end2)/(1024*heads*sectors);
		if(count1024 > 0) {
			end2 = pchs_to_linear(part->endcylinder+count1024*1024,
				part->endhead, part->endsector);
			printf("Warning: end cylinder invalid (cyl >= 1024)\n");
		}
		if(end != end2) 
			printf("Warning: inconsistent end = %d, chs -> %d\n",
				end, end2);
	}
	return part->size;
}



bool
read_sector(fd_t devfd, void *adr, uint32 secnum)
{
	lseek(devfd, secnum*SECTOR_SIZE, SEEK_SET);
	return read(devfd, adr, SECTOR_SIZE) == SECTOR_SIZE;
}


bool
read_sector(fd_t devfd, void *adr, int cyl, int head, int sec)
{
	return read_sector(devfd, adr, pchs_to_linear(cyl, head, sec));
}



/*  Table format:
Part AF   Cyl Hd Sec   Cyl Hd Sec    Start     Size     MB ID Type
 1/a 80 00000 00 000 00000 00 000 _______1 _______4 ___3.0 01 DOS
 2/b -- 00000 00 000 00000 00 000 _____234 ____4545 _455.2 83 Linux
 3/c -- 00000 00 000 00000 00 000 _____234 ____4545 _455.3 0a OS/2 Boot Manager
*/
 
void
print_partition_entry(int index, Partition *part, uint offset = 0)
{
	uint startcylinder = part->startcylinder;
	uint starthead = part->starthead;
	uint startsector = part->startsector;
	uint endcylinder = part->endcylinder;
	uint endhead = part->endhead;
	uint endsector = part->endsector;
	uint size_tenthMB;

	chs_unpack(startcylinder, starthead, startsector);
	chs_unpack(endcylinder, endhead, endsector);
	size_tenthMB = ((uint)part->size*10)/(_M_/SECTOR_SIZE);

	if(index)
		printf("%2d/%c ", index, index-1+'a');
	else
		printf("     ");
	if(part->boot)
		printf("%02x", part->boot);
	else
		printf("--");
	printf(" %05d %02d %03d %05d %02d %03d"
		"%9u%9u%5u.%1u %02x %s\n", 
		startcylinder, starthead, startsector,
		endcylinder, endhead, endsector,
		(uint)part->start+offset, (uint)part->size,
		size_tenthMB/10, size_tenthMB%10,
		part->id, ptype_name(part->id));
}

 
void
print_partition_table(fd_t devfd)
{
	BootBlock sec;
	int ext, extofs;
	int lognum;
	int start;
	int size;

	printf("Part AF   Cyl Hd Sec   Cyl Hd Sec    "
			"Start     Size     MB ID Type\n");
	if(!read_sector(devfd, &sec, 0)) {
		printf("error: could not read boot sector\n");
		return;
	}
	ext = 0;
	for(int i=0; i<4; i++) 
		if(sec.partition[i].id) {
			print_partition_entry(i+1, &sec.partition[i]);
			start = partition_start(&sec.partition[i]);
			size = partition_size(&sec.partition[i]);
			if(sec.partition[i].id == PARTID_EXTENDED
			|| sec.partition[i].id == 0xf	/* windows ext */
			|| sec.partition[i].id == 0x85)	/* linux ext */
				ext = start;
		}
	lognum = 4;
	extofs = ext;
	while(ext) {
		if(!read_sector(devfd, &sec, ext)) {
			printf("error: could not read sector %d\n", ext);
			return;
		}
		print_partition_entry(++lognum, &sec.partition[0], ext);
		start = partition_start(&sec.partition[0], ext);
		size = partition_size(&sec.partition[0], ext);
		ext = 0;
		if(sec.partition[1].id == PARTID_EXTENDED
		|| sec.partition[1].id == 0xf	/* windows ext */
		|| sec.partition[1].id == 0x85)	/* linux ext */
		{
			print_partition_entry(0, &sec.partition[1], extofs);
			ext = partition_start(&sec.partition[1], extofs);
		}
		// for debugging...
		// print_partition_entry(lognum, &sec.partition[1]);
	}
}


sz*
ide_info_strncpy(sz *dest, csz *sp, int n)
{
	sz *dp = dest;
	csz *ep = sp + n - 1;
	int bswap = 1;

	int space = 0;
	while(sp[bswap] && sp < ep) {
		char ch = (sp++)[bswap];
		bswap = -bswap;
		if(ch != ' ') {
			*dp++ = ch;
			space = 1;
		} else if(sp < ep && sp[bswap] != ' ') {
			*dp = ' ';
			dp += space;
		}
	}
	*dp = 0;
	return dest;
}


void
get_ide_drive_info(int drive)
{
	ushort id[256];
	sz buf[50];

	printf("Controller %d, Drive %d:\n", (drive&2)>>1, drive&1);

#ifdef USE_DEV_PORT  // outdated code...
	fd_t port = open("/dev/port", O_RDWR);
	fd_t rportw = open("/dev/rportw", O_RDONLY);
	byte cmd[2];

	if(port < 0) {
		printf("could not open /dev/port\n");
		return;
	}
	if(rportw < 0) {
		printf("could not open /dev/rportw\n");
		return;
	}
	cmd[0] = 0xa0+((drive&1) << 4);	// select drive 'drive'
	cmd[1] = 0xec;			// Indentify drive
	lseek(port, 0x1f6, SEEK_SET);
	write(port, cmd, 2);
	do {
		lseek(port, 0x1f7, SEEK_SET);
		read(port, cmd, 1);
	} while((cmd[0] & 0xc8) != 0x48);
	lseek(rportw, 0x1f0, SEEK_SET);
	read(rportw, id, sizeof(id));
	close(port);
	close(rportw);
#else
	int disk_iobase = (drive&2) ? 0x170 : 0x1f0;

	Port16 hd_data(disk_iobase+0);
	Port8 hd_select(disk_iobase+6);
	Port8 hd_cmd(disk_iobase+7);
	Port8 hd_status(disk_iobase+7);

	hd_select = 0xa0+((drive&1) << 4); // select drive 
	if((hd_status & 0xc0) != 0x40) {
		printf("Drive not responding\n");
		return;
	}

	hd_cmd = 0xec;		 // Get Drive Information
	while((hd_status & 0xc0) != 0x40)
		continue;

	if(hd_status & 1) {
		printf("GetDriveInformation not supported by this drive\n");
		return;
	}

	for(size_t i=0; i<nelem(id); i++)
		id[i] = hd_data;
#endif


	printf("Model: %s, ", ide_info_strncpy(buf, (csz*)&id[27], 41));
	printf("Revision: %s, ", ide_info_strncpy(buf, (csz*)&id[23], 9));
	printf("SN: %s\n", ide_info_strncpy(buf, (csz*)&id[10], 21));

	printf( "Buffer size: %dK, "
		"Configuration: %04x\n", id[21]/2, id[0]);

	printf("  ");
	if(id[0] & 0x40) printf("fixed drive, ");
	if(id[0] & 0x80) printf("removable media, ");
	if(id[0] & 0x02) printf("hard-sectored, ");
	if(id[0] & 0x04) printf("soft-sectored, ");
	if(id[0] & 0x08) printf("RLL/ARLL format\n");
		else printf("MFM encoded\n");
	if(id[0] & 0x10) printf("  head switch < 15ms, ");
		else printf("  head switch >= 15ms, ");
	if(id[0] & 0x20) printf("  power down supported\n");
	if(id[0] & 0x100) printf("  data rate < 5Mb/s\n");
	if(id[0] & 0x200) printf("  data rate between 5Mb/s and 10Mb/s\n");
	if(id[0] & 0x400) printf("  data rate > 10Mb/s\n");
	if(id[0] & 0x800) printf("  rotation deviation > 0.5%%\n");
	printf("  buffer type %04x: ", id[20]);
	switch(id[20]) {
	case 1: printf("single-sector\n"); break;
	case 2: printf("multi-sector, dual-ported\n"); break;
	case 3: printf("multi-sector, dual-ported, w/ read cache\n"); break;
	default: printf("unknown\n"); break;
	}
	if(id[48] & 1) printf("  32-bit I/O, ");
	else printf("  16-bit I/O, "); 

	printf("maxmult=%d, ", id[47]&0xff);
	printf("%s DMA, %s LBA, %s logical, %s EIDE\n",
		(id[49]&1) ? "" : "no",
		(id[49]&2) ? "" : "no",
		(id[53]&1) ? "" : "no",
		(id[53]&2) ? "" : "no");
	printf("  PIO cycle time: ");
	switch(id[51]>>8) {
	case 0: printf("600 ns"); break;
	case 1: printf("380 ns"); break;
	case 2: printf("240 ns"); break;
	case 3: printf("180 ns"); break;
	default: printf("unknown"); break;
	}
	printf(", DMA cycle time: ");
	switch(id[51]>>8) {
	case 0: printf("960 ns"); break;
	case 1: printf("380 ns"); break;
	case 2: printf("240 ns"); break;
	case 3: printf("150 ns"); break;
	default: printf("unknown"); break;
	}
	printf("\n");

	int secsize = id[5];
	if(secsize > 512 && secsize < 1024)
		secsize = 512;
	if(secsize > 1024 && secsize < 2048)
		secsize = 1024;
	printf("  unformatted bytes/sector: %d  bytes/track: %d\n",
		id[5], id[4]);
	printf("Physical geometry: c:%d h:%d s:%d ss:%d t:%dMb\n",	
		id[1], id[3], id[6], secsize, 
		(id[1]*id[3]*id[6])/(_M_/secsize));
	printf("Logical geometry: c:%d h:%d s:%d ts:%d t:%dMb\n",
		id[54], id[55], id[56], (id[58]<<16)|id[57],
		((id[58]<<16)|id[57])/(_M_/secsize));
	printf("LBA: ts:%d t:%d\n",
		(id[60]<<16)|id[61],
		((id[60]<<16)|id[61])/(_M_/secsize));
}


int
main(int argc, csz **argv)
{
	csz *devname = "/dev/hda";
	fd_t devfd;
	FileInfo devinfo;

	/* This note isn't needed for now, since we can't write anyways...
	printf("NOTE: make sure you know what you're doing before using\n"
	       "this program.  This utility could DESTROY YOUR ENTIRE HARD DISK\n"
	       "if used improperly, or if there is a bug in it, so make sure\n"
	       "all your important data is backed up before proceeding!\n"
	       "Enter 'q' to simply quit if you're not sure.\n"
	       "\n");
	printf("Note2: This program is currently incomplete, and is missing\n"
	       "most of its necessary functionality.  Use a standard FDISK\n"
	       "program instead for now.\n"
	       "\n");
	*/
	for(int i=1; i<argc; i++) {
		if(!strcmp(argv[i], "-v")) {
			printf("fdisk " fdisk_VER_STR "\n");
			exit(0);
		} else
			devname = argv[i];
	}
	if((devfd = open(devname, O_RDWR)) < 0) {
		printf("could not open %s\n", devname);
		exit(1);
	}
	solo_fstat(devfd, &devinfo);
	cylinders = devinfo.cylinders;
	heads = devinfo.heads;
	sectors = devinfo.sectors;
	while(true) {
		printf("Command (m for help): ");
		sz buf[80];
		while(!gets(buf, sizeof(buf))) continue;
		switch(buf[0]) {
		case 'I':
			get_ide_drive_info(RDEV_UNIT(devinfo.rdev) & 0xf);
			break;
		case 'p':
			printf("Disk %s: %d cylinders, %d heads, %d sectors\n\n",
				devname, cylinders, heads, sectors);
			print_partition_table(devfd);
			break;
		case 'c':
			printf("Number of cylinders (1-65535, currently %d): ",
					cylinders);
			cylinders = getint(cylinders);
			break;
		case 'h':
			printf("Number of heads (1-63, currently %d): ",
					heads);
			heads = getint(heads);
			break;
		case 's':
			printf("Number of sectors (1-255, currently %d): ",
					sectors);
			sectors = getint(sectors);
			break;
	        case 'w':
			printf("saving changes\n");
		case 'q':
			printf("exiting...\n");
			close(devfd);
			exit(0);
		default:
			printf("Command action\n"
			       "   m   print this menu\n"
			       "   I   get IDE drive info\n"
			       "   p   print partition table\n"
			       "   c   change number of cylinders\n"
			       "   h   change number of heads\n"
			       "   s   change number of sectors\n"
			       "   n   create new partition (unimplemented)\n"
			       "   q   quit without saving changes\n"
			       "   w   write changes and exit\n");
		}

	} 
}

