/*  $Id: soloext.cc,v 1.8 2000/01/12 03:44:47 shaggy Exp $
 *      Extensions to SOLO -- "builtin" commands
 *      relative path: sys/boot/m-pc
 *  
 *  Copyright (c) 1999 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/members/shaggy
 */

#define soloext_VER_MAJOR	0
#define soloext_VER_MINOR	1
#define soloext_VER_PATCH	0
#define soloext_VER_STR	"0.1"
#define soloext_VER_DATE	"June 20, 1999"
#define soloext_VER_NDATE	"06/20/99"
 
#include <libc.h>
#include <stdx1.h>
#include <defines>
#include <boot/bootparam.h>
#include <boot/cpuinfo.h>
#include <boot/regblock.h>
#include <boot/video.h>
#include <hw/Port.h>
#include "bios.h"
#include "vmem.h"


extern BootParam *bootparam;
extern "C" const csz **environ;


static csz rcsid[] __attribute__((unused))
	= "$Id: soloext.cc,v 1.8 2000/01/12 03:44:47 shaggy Exp $"
	  "$Release: " soloext_VER_STR " $";

void maplo4k()
{
	ulong *pd;
	asm("movl %%cr3,%0":"=r"(pd));
	pd = (ulong*)(pd[0]&~(PAGE_SIZE-1));
	pd[0] = 0x00000007;	/* phys 0, user RW, present */
}

ushort*
vidbase()
{
	Port(0x3ce).outb_p(6);
	if(Port(0x3cf).inb_p() & 1)
		return (ushort*)0xa0000;
	if(Port(0x3cc).inb_p() & 1)
		return (ushort*)0xb8000;
	else
		return (ushort*)0xb0000;
}

int savex;
int savey;
ushort *savescr;

void
vidsave()
{
	ushort *v = (ushort*)vidbase();
	if(v == (ushort*)0xa0000) {
		savescr = 0;
		return;
	}
	savex = *(byte*)BIOS_VID_X;
	savey = *(byte*)BIOS_VID_Y;
	int wid = *(byte*)BIOS_VID_WID;
	int hgt = *(byte*)BIOS_VID_HGT+1;
	ushort *s = (ushort*)malloc(2*256*hgt);
	memsetw(s, 0x0720, 256*hgt);
	savescr = s;
	for(int y=0; y<hgt; y++) {
		memcpy(s, v, wid*2);
		s += 256;
		v += wid;
	}
}

void
vidrestore()
{
	ushort *v = (ushort*)vidbase();
	if(v == (ushort*)0xa0000)
		return;
	int wid = *(byte*)BIOS_VID_WID;
	int hgt = *(byte*)BIOS_VID_HGT+1;
	ushort *s = savescr;
	if(!savescr)
		return;
	if(savey >= hgt) {
		s += 256*(savey-hgt+1);
		savey = hgt-1;
	} else
		hgt = savey+1;
	if(savex >= wid)
		savex = wid-1;
	for(int y=0; y<hgt; y++) {
		memcpy(v, s, wid*2);
		s += 256;
		v += wid;
	}
	*(byte*)BIOS_VID_X = savex;
	*(byte*)BIOS_VID_Y = savey;
	free(savescr);
}


void
normalize(void **p)
{
	ulong a = (ulong)*p;
	a = ((a>>12)&0x000ffff0)+(a&0xffff);
	*p = (void*)a;
}

void
vinit()
{
	RegBlock reg;
	VidInfo *vi = bootparam->vidmode->vidinfo;
	reg.eflags = 0;
	reg.eax = 0x4f00;
	reg.edi = (ulong)vi;
	memset(vi, 0, sizeof *vi);
	bootparam->callbios(0x10, &reg);
	if((reg.eax&0xff) != 0x4f) {
		printf("(warning: no VESA BIOS detected)\n");
		return;
	}
	memmove(&vi->_align+1, &vi->_align,
			(ulong)(vi+1) - (ulong)(&vi->_align+1));
	normalize((void**)&vi->oemstring);
	normalize((void**)&vi->modelist);
	normalize((void**)&vi->oemvendor);
	normalize((void**)&vi->oemproduct);
	normalize((void**)&vi->oemproductrev);
}

int vmode(int mode);

int
vquery(struct VidMode *vm, int mode)
{
	RegBlock reg;
	VidInfo *vi = bootparam->vidmode->vidinfo;
	int curmode = bootparam->vidmode->mode;
	VidMode *pvm = ((ulong)vm >= 0x00100000)
		? (VidMode*)(bootparam->buffer+0x100) : vm;
	reg.eflags = 0;
	reg.eax = 0x4f01;
	reg.ecx = mode;
	reg.edi = (ulong)pvm;
	memset(pvm, 0, sizeof *pvm);
	bootparam->callbios(0x10, &reg);
	if(pvm != vm)
		memmove(vm, pvm, sizeof *pvm);
	vm->vidinfo = vi;
	vm->mode = mode;
	if(mode <= 0xff && (reg.eax&0xffff) == 0x014f)
		reg.eax = 0;
	if((reg.eax&0xff) != 0x4f) {
		if(mode > 0xff) 
			return -1;
		int m = mode;
		if(mode != curmode) {
			vidsave();
			reg.eax = mode;
			bootparam->callbios(0x10, &reg);
			m = *(byte*)BIOS_VID_MODE;
		}
		vm->xres = *(byte*)BIOS_VID_WID;
		vm->yres = *(byte*)BIOS_VID_HGT+1;
		if(vm->xres < 2 || vm->yres < 2)
			m = -1;
		if(m != mode) {
			if(m != curmode) {
				reg.eax = curmode;
				bootparam->callbios(0x10, &reg);
			}
			if(mode != curmode)
				vidrestore();
			return -1;
		}

		if(Port(0x3c5).inb_p() & 1)
			vm->xchar = 8;
		else
			vm->xchar = 9;
		vm->ychar = *(byte*)BIOS_VID_BPCHAR;
		ushort *v = vidbase();
		vm->vidinfo = vi;
		vm->winaseg = (ulong)v>>4;
		vm->attr = VidMode::ATTR_SUPPORTED|VidMode::ATTR_TTYOUT;
		if(v == (ushort*)0xb8000) {
			vm->attr |= VidMode::ATTR_COLOR;
			vm->memmodel = VidMode::MMOD_TEXT;
		} else if(v == (ushort*)0xb0000) {
			vm->memmodel = VidMode::MMOD_TEXT;
		} else {
			vm->attr |= VidMode::ATTR_COLOR|VidMode::ATTR_GRAPHICS;
			vm->xres *= vm->xchar;
			vm->yres *= vm->ychar;
			reg.eax = 0x1b00;
			reg.ebx = 0x0000;
			reg.edi = bootparam->buffer+0x100;
			bootparam->callbios(0x10, &reg);
			if((reg.eax&0xff) == 0x1b) {
				uint n = *(byte*)(reg.edi+0x27)
					+((*(byte*)(reg.edi+0x28))<<8);
				while(n > 1) {
					vm->depth++;
					n >>= 1;
				}
			}
			// XXX - needs fixing:
			if(vm->depth == 8)
				vm->memmodel = VidMode::MMOD_PACKED;
			else if(vm->depth == 4)
				vm->memmodel = VidMode::MMOD_PLANAR;
			else
				vm->memmodel = VidMode::MMOD_CGA;
		}

		if(mode != curmode) {
			reg.eax = curmode;
			bootparam->callbios(0x10, &reg);
			vidrestore();
		}
	} else {
		if((reg.eax&0xff00) != 0) 
			return -1;
		memmove(&vm->_align+1, &vm->_align,
			(ulong)(vm+1) - (ulong)(&vm->_align+1));
	}
	if(!(vm->attr&VidMode::ATTR_GRAPHICS)
			 && (vm->xres >= 640 || vm->yres >= 400)) {
		// the VESA 3.0 spec says that in text modes, xres and yres
		// are the number of character columns and rows, but some
		// boards (like the ATI Expert 128) still put the pixel
		// resolution here, so we need to check for it and
		// convert it

		vm->xres /= vm->xchar;
		vm->yres = (vm->yres+1)/vm->ychar;
	}
	if(mode == curmode)
		vm->crtc = *(ushort*)BIOS_VID_CRTC;
	return 1;
}

int
vmode(int mode)
{
	RegBlock reg;
	VidMode *vm = bootparam->vidmode;
	int oldmode = vm->mode;
	reg.eflags = 0;
	reg.eax = 0x4f02;
	reg.ebx = mode;
	bootparam->callbios(0x10, &reg);
	if((reg.eax&0xff) != 0x4f) {
		if(mode > 0xff)
			return -1; 
		reg.eax = mode;
		bootparam->callbios(0x10, &reg);
	} else {
		if((reg.eax&0xff00) != 0)
			return -1;
	}
	vm->mode = mode;
	if(vquery(vm, mode) < 0) {
		vmode(oldmode);
		return -1;
	}
	return 0;
}

int
vmode(int argc, csz *argv[])
{
	maplo4k();
	vinit();
	if(argc > 1) {
		int m = strtol(argv[1], 0, 16);
		if(m < 0)
			vquery(bootparam->vidmode, bootparam->vidmode->mode);
		else {
			vidsave();
			vmode(m);
			vidrestore();
		}
	} else {
		VidMode *vm = bootparam->vidmode;
		printf("%04x: %dx%dx%d (%dx%d) attr=%x wina=%x,%x winb=%x,%x gran=%x size=%x span=%d\n",
			vm->mode, vm->xres, vm->yres, vm->depth,
			vm->xchar, vm->ychar, vm->attr,
			vm->winaattr, vm->winaseg,
			vm->winbattr, vm->winbseg,
			vm->wingran, vm->winsize,
			vm->span);
		printf("\tbase=%lx memmodel=%x nplanes=%d nbanks=%d banksize=%x nimage=%d\n", 
			vm->base, vm->memmodel, vm->nplanes,
			vm->nbanks, vm->banksize, vm->nimage);
		printf("\tlinspan=%d linnimage=%d maxclock=%ld crtc=%x\n",
			vm->linspan, vm->linnimage, vm->maxpixclock, vm->crtc);
		VidInfo *vi = vm->vidinfo;
		printf("%c%c%c%c %x.%x \"%s\" %dKB cap:%lx\n",
			vi->sig[0], vi->sig[1], vi->sig[2], vi->sig[3],
			vi->vers>>8, vi->vers&0xff,
			vi->oemstring,
			vi->totmem*64,
			vi->cap);
		if(vi->oemvendor)
			printf("\t\%x \"%s\" \"%s\" \"%s\"\n",
				vi->oemsoftrev,
				vi->oemvendor,
				vi->oemproduct,
				vi->oemproductrev);
	}
	return 0;
}


int
vscan(int argc, csz *argv[])
{
	maplo4k();
	vinit();

	int matchx = 0;
	int matchy = 0;
	int matchd = 0;
	ushort amask = VidMode::ATTR_SUPPORTED;
	ushort aval = VidMode::ATTR_SUPPORTED;
	int m = 0;
	int mtop = 0x1ff;
	VidMode vm;
	VidInfo *vi = bootparam->vidmode->vidinfo;
	ushort *ml = vi->modelist;
	int c = 0;

	++argv; --argc;
	while(argc > 0) {
		if((*argv)[0] != '-')
			break;
		char c = (*argv)[1];
		++argv; --argc;
		if(argc < 1)
			return 1;
		int n = strtol(*argv, 0, 0);
		++argv; --argc;
		switch(c) {
		case 'a': amask |= n; aval |= n; break;
		case 'n': amask |= n; aval &= ~n; break;
		case 'x': matchx = n; break;
		case 'y': matchy = n; break;
		case 'd': matchd = n; break;
		default:
			return 1;
		}
	}
	if(argc > 0) {
		m = strtol(*argv, 0, 16);
		++argv; --argc;
		ml = 0;
	}
	if(argc > 0) 
		mtop = strtol(*argv, 0, 16);

	if(ml && *ml == 0xffff) 
		ml = 0;

	for(;;) {
		sz buf[39];
		if(ml) {
			m = *ml++;
			if(m == 0xffff)
				break;
		} else if(m > mtop)
			break;
		if(vquery(&vm, m) >= 0 && (vm.attr&amask) == aval
				&& (!matchx || vm.xres == matchx)
				&& (!matchy || vm.yres == matchy)
				&& (!matchd || vm.depth == matchd)) {
			sprintf(buf, "%x:           (%dx%d,%04x,%x) %dx%d",
				m, vm.xchar, vm.ychar, vm.attr, vm.memmodel,
				vm.xres, vm.yres);
			if(vm.attr&VidMode::ATTR_GRAPHICS)
				sprintf(buf, "%x: %dx%dx%d (%dx%d,%04x,%x) %dx%d",
					m, vm.xres, vm.yres, vm.depth,
					vm.xchar, vm.ychar, vm.attr, vm.memmodel,
					vm.xres/vm.xchar, vm.yres/vm.ychar);
			c += sizeof buf;
			if(c >= 80) {
				printf("\n");
				c = sizeof buf; 
			}
			printf("%-39s", buf);
		}
		++m;
		if(m > 0x7f && m < 0x100)
			m = 0x100;
	}
	if(c > 0)
		printf("\n");
	else {
		printf("No matching video modes found\n");
		return 1;
	}
	return 0;
}


int unzip(int argc, csz *argv[]);
int lsboot(int argc, csz *argv[]);

struct {
	csz *name;
	int (*func)(int argc, csz *argv[]);
} cmdtab[] = {
	{"vscan", vscan},
	{"vmode", vmode},
	{"unzip", unzip},
	{"lsboot", lsboot},
};


int
main(int argc, csz *argv[])
{
	for(uint i=0; i<nelem(cmdtab); i++) {
		if(strcmp(argv[0], cmdtab[i].name) == 0) {
			cmdtab[i].func(argc, argv);
			return 0;
		}
	}
	return 127;	// command not found
}

