/*  $Id: solotest.cc,v 1.21 1999/07/16 20:18:51 shaggy Exp $
 *      Test SOLO BootParam interface
 *      relative path: ./sys/boot/m-pc
 */ 

#define solotest_VER_MAJOR	0
#define solotest_VER_MINOR	97
#define solotest_VER_PATCH	0
#define solotest_VER_STR	"0.97"
#define solotest_VER_DATE	"January 20, 1997"
#define solotest_VER_NDATE	"01/20/97"

#include <libc_posix.h> 
#include <stdx1.h>
#include <defines>
#include <boot/bootparam.h>
#include <boot/cpuinfo.h>
#include <boot/regblock.h>
#include <hw/Port.h>

extern BootParam *bootparam;
extern "C" csz **environ;


static csz rcsid[] __attribute__((unused))
	= "$Id: solotest.cc,v 1.21 1999/07/16 20:18:51 shaggy Exp $"
	  "$Release: " solotest_VER_STR " $";



void
show_file(csz *fname, int bufsize)
{
	fd_t fd;
	sz buf[bufsize];
	int n;

	printf("opening \"%s\"\n", fname);
	if((fd = open(fname, O_RDONLY)) < 0) {
		printf("no such file\n");
		return;
	}
	printf("fd=%x\n", fd);
	while((n = read(fd, buf, bufsize))) 
		write(STDOUT_FILENO, buf, n);
	printf("total bytes read: %ld\n", lseek(fd, 0, SEEK_CUR));
	close(fd);
}


void
raw_con_test()
{
	fd_t raw_con;
	uchar ch;

	if((raw_con = open("/dev/rawcon", O_RDONLY)) < 0) {
		printf("cannot open /dev/rawcon\n");
		return;
	}
	printf("Press any key, or 'q' to quit\n");
	do {
		read(raw_con, &ch, 1);
		int n = ch;
		if(ch < 32) ch = '.';
		printf("you pressed key %d/0x%02x (%c)\n", n, n, ch);
	} while(ch != 'q');
	close(raw_con);
}



void
mini_debug(sz *cmdstr)
{
	sz *cmd;
 
	cmd = strtok(cmdstr, " ");
	while(cmd) {
		bool port = false;
		if(*cmd == '%') {
			port = true;
			cmd++;
		}
	
		sz *pos = strchr(cmdstr, '=');
		if(pos) {
			*pos++ = 0;
			uint32 adr = strtol(cmd, 0, 0);
			uint32 val = strtol(pos, 0, 0);
			if(port)
				(Port8(adr)) = val;
			else
				*(uint32*)adr = val;
		} else {
			uint32 adr = strtol(cmd, 0, 0);
			printf("0x%lx\n", 
				port ?  (uint32)Port8(adr) : *(uint32*)adr);
		}
		
		cmd = strtok(0, " ");
	}
}


int
main(int argc, csz **argv, csz **envp)
{
	int bufsize = 4096;
	int x;
	sz buf[80];

	printf("SOLO Test for SOLO v%d.%d.X\n", solotest_VER_MAJOR,
						solotest_VER_MINOR);  
	buf[0] = '\0';

	while(true) {
		if(!buf[0]) {
			printf("  1) Print BootParam info\n");
			printf("  2) Print Arguments\n");
			printf("  3) Print Environment\n");
			printf("  4) Test console input\n");
			printf("  5) Test raw console input\n");
			printf("  6) Test file reading (bufsize=%d)\n", bufsize);
			printf("  7) Change buffer size\n");
			printf("  8) Test panic\n");
			printf("  9) Test exceptions/interrupts\n");
			printf(" 10) Libsolo tests\n");
			printf(" 12) Test BIOS calls\n");
			printf("  d) Mini debugger command\n");
			printf("  q) Quit\n");
		}

	get_command:
		printf("Enter option: ");
		if(!gets(buf, sizeof(buf)) || buf[0] == 'q')
			break;
		if(buf[0] == 'd') {
			mini_debug(&buf[1]);
			goto get_command;
		}

		printf("\n");
		switch(atoi(buf)) {
		case 1:
			printf("Bootparam = %p\n", bootparam);
			printf("Boot loader: %s v%d.%d (dev=0x%02x)\n",
				bootparam->bootname, 
				bootparam->bootver_major,
				bootparam->bootver_minor,
				bootparam->bootdev);
			printf("argc=%ld argv=%p envp=%p\n",
				bootparam->argc, bootparam->argv,
				bootparam->envp);
			printf("BIOS call buffer: %lx-%lx (%d bytes)\n",
				bootparam->buffer, 
				bootparam->buffer+bootparam->bufsize-1,
				bootparam->bufsize);
			printf("Entry=%lx lomem=%lx himem=%lx\n",
				bootparam->entry, bootparam->lomem,
				bootparam->himem);
			printf("Base memory: %lx-%lx (%ldK)\n",
				bootparam->base_addr, 
				bootparam->base_addr+bootparam->base_size-1,
				bootparam->base_size/1024);
			printf("base_lomem=%lx base_himem=%lx\n",
				bootparam->base_lomem, bootparam->base_himem);
			printf("Extended memory: %lx-%lx (%ldK)\n",
				bootparam->ext_addr, 
				bootparam->ext_addr+bootparam->ext_size-1,
				bootparam->ext_size/1024);
			printf("Msgbuf=%lx msgsize=%x msgptr=%x\n",
				bootparam->msgbuf, bootparam->msgsize,
				bootparam->msgptr);
			printf("Cpu: family=%d name=%s vendor=%s\n"
			       "     model=%d revision=%d type=%d flags=%08lx\n",
				bootparam->cpuinfo->family,
				bootparam->cpuinfo->name,
				bootparam->cpuinfo->vendor,
				bootparam->cpuinfo->model,
				bootparam->cpuinfo->revision,
				bootparam->cpuinfo->type,
				bootparam->cpuinfo->flags);
			break;
		case 2:
			printf("argc=%d argv=%p\n", argc, argv);
			for(int i=0; i<argc; i++)
				printf("argv[%d] = %p:\"%s\"\n", i,
							argv[i], argv[i]);
			break;
		case 3:
			printf("envp=%p environ=%p\n", envp, environ);
			for(csz **ep = environ; *ep; ep++)
				printf("%p: %s\n", *ep, *ep);
			break;
		case 4:
			printf("Enter test string: ");
			if(gets(buf, sizeof(buf)))
				printf("got (%s)\n", buf);
			else
				printf("got EOF\n");
			break;
		case 5:
			raw_con_test();
			break;
		case 6:
			printf("Enter name of file to display: ");
			gets(buf, sizeof(buf));
			show_file(buf, bufsize);
			break;
		case 7:
			printf("Buffer size: ");
			gets(buf, sizeof(buf));
			bufsize = atoi(buf);
			break;
		case 8: {
			printf("Panic type (1=irq, 2=trap, 3=reset, 4=misc): ");
			gets(buf, sizeof(buf));
			int ptype = atoi(buf);
			printf("Panic number (irq/trap/misc #): ");
			gets(buf, sizeof(buf));
			int pnum = atoi(buf);
			printf("Extended error code: ");
			gets(buf, sizeof(buf));
			int pcode = atoi(buf);
			bootparam->panic(ptype, pnum, pcode, 0); 
			} break;
		case 9:
			printf("  0) Null pointer\n");
			printf("  1) Page fault\n");
			printf("  2) Division by 0\n");
			printf("  3) INT 3 (breakpoint)\n");
			printf("  4) IRQ\n");
			printf("  5) CPU reset\n");
			printf("Enter option: ");
			
			gets(buf, sizeof(buf));
			x = 0;
			switch(atoi(buf)) {
			case 0:
				printf("got byte: %d (?)\n", *(uint8*)0x0);
				break;
			case 1:
				printf("got byte: %d (?)\n", *(uint8*)0xdeadbeef);
				break;
			case 2:
				printf("answer=%d (?)\n", 123/x);
				break;
			case 3:
				asm("int $3");
				printf("returned from breakpoint (?)\n");
				break;
			case 4:
				printf("Turning on interrupts, enabling mask 0x0101,\n"
				       "and going into infinite loop...\n"
				       "press any key to cause an interrupt\n");
				asm("movb $1,%al; outb %al,$0x21; outb %al,$0xa1; sti");
				while(true) continue;
			case 5:
				asm("lidt 1f; int $3; 1: .long 0, 0");
				printf("(?)\n");
				break;
			}
			break;
		case 10:
			printf("The following commands are available:\n");
			printf("	m <size>	malloc(size)\n");
			printf("	r <adr> <size>	realloc(adr,size)\n");
			printf("	f <adr>		free(adr)\n");
			printf("	g <str>		getenv(str)\n");
			printf("	q		quit\n");
			while(true) {
				printf("cmd: ");
				if(!gets(buf, sizeof(buf)) || buf[0] == 'q')
					break;
				sz *cp = &buf[1];
				uint32 n1 = strtoul(cp, &cp, 0);
				uint32 n2 = strtoul(cp, &cp, 0);
				sz *str = &buf[1];
				while(*str && *str == ' ')
					str++;
				switch(buf[0]) {
				case 'm':
					printf("%p\n", malloc(n1));
					break;
				case 'r':
					printf("%p\n", realloc((void*)n1, n2));
					break;
				case 'f':
					free((void*)n1);
					break;
				case 'g':
					printf("getenv(\"%s\") = \"%s\"\n",
						str, getenv(str));
					break;
				}
			}
			break;
		case 11:
			printf("This option intentionally left blank\n");
			break;
		case 12:
			RegBlock reg;

			printf("  0) Change video mode (int 10)\n");
			printf("  1) Get KB character (int 16)\n");
			printf("Enter option: ");
			
			gets(buf, sizeof(buf));
			switch(atoi(buf)) {
			case 0:
				reg.eax = 0x0002;
				bootparam->callbios(0x10, &reg);
				break;
			case 1:
				reg.eax = 0x0000;
				bootparam->callbios(0x16, &reg);
				printf("key=%#lx\n", reg.eax);
				break;
			default:
				printf("invalid option\n");
			}
			break;			
		default:
			printf("Invalid option\n");
		}
		printf("\n");
	} 
	printf("exiting\n");
}

